package Gtk::HexEntry;
our @ISA = qw(Gtk::Entry);
use Data::Dumper;

sub new {
	my $obclass = shift;
	my $class = ref($obclass) || $obclass;
	my $adj = shift;  # adjustment

	my $self = $class->SUPER::new();
	bless $self, $class;

	$self->signal_connect( 'focus_in_event',  sub { $self->select_region(0, -1); return 0; } );
	$self->signal_connect( 'focus_out_event', sub { $self->focus_out(@_);  } );
	$self->signal_connect( 'activate',        sub { $self->update_value(@_); } );
	$adj->signal_connect( 'value_changed', sub { $self->value_changed(@_) } );

	$self->{adj} = $adj;
	$self->value_changed;
	
	return $self;
}

sub set_value {
	my $self = shift;
	my $val  = shift;
	$self->{adj}->set_value($val) if defined $val;
	return $val;
}

sub value_changed {
	my $self = shift;
	$self->set_text( sprintf("%x", $self->{adj}->value()) );
	return 1;
}

sub update_value {
	my $self = shift;
	$self->{adj}->set_value(hex $self->get_text());
	# Note: set_value will not emit value_changed if the values are the same.
	# However, we must always call value_changed to convert to hex.
	# We could call value() and emit value_changed by hand, but then we'd also
	# have to clamp by hand.  So we could call value_changed twice, so what.
	$self->value_changed();
	return 1;
}

sub value {
	my $self = shift;
	return $self->{adj}->value;
}

sub focus_out {
	my $self = shift;
	$self->update_value(); 
	$self->select_region(0, 0); 
	return 1;
}

1;

