package U6::BookEdit;
use U6::Book qw(@books);

use strict;

# $adj = new Gtk::Adjustment( $value, $lower, $upper, $step_increment, $page_increment, $page_size );
# $spin = new Gtk::SpinButton( $adjustment, $climb_rate, $digits );
# $text = new Gtk::Text( $hadj, $vadj );

sub create {
	my $obclass = shift;
	my $class = ref($obclass) || $obclass;
	my $adj  = new Gtk::Adjustment( -1, -1, $#books, 1, 10, 0 );
	my $self = {
		# spin and win!
		SPIN => new Gtk::SpinButton( $adj, 1.0, 0),
		WIN  => new Gtk::Dialog(),
	};
	bless $self, $class;

	my $text = new Gtk::Text( undef, undef );
	my $text_hbox = new Gtk::HBox(0, 0);
	my $text_vscroll = new Gtk::VScrollbar( $text->vadj );

	$text->set_editable( 1 );
	$text->set_word_wrap( 1 );
	$text->set_line_wrap( 1 );
	my $color_red = $text->get_colormap->color_alloc( { red => 0xffff, green => 0, blue => 0 } );

	$text_hbox->pack_start($text, 1, 1, 0);
	$text_hbox->pack_start($text_vscroll, 0, 0, 0);

	signal_connect $adj 'value_changed', sub { 
		my $val = $_[1]->get_value_as_int; 
		my $text = $_[2];
		$text->set_point(0);
		$text->forward_delete( $text->get_length );  # faster way to clear text box?
	#	$text->insert( undef, $color_red, undef, "book follows\n");
		if ($val >= 0) {
			$text->insert( undef, undef, undef, $books[$val]); 
		} else {
			# Negative value: invalid (empty) book
			$text->insert( undef, undef, undef, "[This book has no contents.]");
		}
	}, $self->{SPIN}, $text;

	# Force an update of the book text.
	$adj->signal_emit('value_changed');

	$self->{WIN}->vbox->pack_start($self->{SPIN}, 0, 0, 0);
	$self->{WIN}->vbox->pack_start($text_hbox, 1, 1, 0);
	$self->{WIN}->set_title( "Book Editor" );

	$self->{WIN}->signal_connect('destroy', sub {
		undef $self->{WIN};
	});

	return $self;
}

sub book {
	my $self = shift;
	return undef unless $self->{WIN};
	$self->{SPIN}->set_value(shift) if @_;
	return $self->{SPIN}->get_value_as_int;
}

# The current semantics for show are: show existing window if available,
# otherwise create a new one and show it.  Other operations (such as book)
# are refused unless the window exists.  This should probably be taken
# care of by a subclass or manager instead.
sub show {
	my $self = shift;
	$self = $self->create() unless $self->{WIN};
	$self->{WIN}->show_all();
	return $self;
}

1;
