#!/usr/bin/perl -w
# $Id: Map.pm,v 1.6 2003/02/12 23:58:53 jim Exp $

package U6::Map;
BEGIN {
	use Exporter;
	our ($VERSION, @ISA, @EXPORT, @EXPORT_OK);
	$VERSION = do { my @r = (q$Revision: 1.6 $ =~ /\d+/g); sprintf "%d."."%02d" x $#r, @r };
	@ISA = qw(Exporter);
	@EXPORT    = qw();
	@EXPORT_OK = qw(@map @chunks chunk_to_world world_to_chunk maptile_at);
}

use strict;

our @map;
our @chunks;

my %filename = ( chunks => "files/chunks", mapfile => "files/map" );

sub init {
	local $/;
	open(CHUNKS, $filename{chunks}) or die "$!";
	@chunks = ();
	while (1) {
		my ($chk, @chk);
		last unless read CHUNKS, $chk, 64;
		@chk = unpack("C*", $chk);
		push @chunks, \@chk;
	}
	close (CHUNKS);

	open(MAP, $filename{mapfile}) or die "$!";
	@map = ();
	# map format: 3 nibbles (little-endian) == chunk #
	# surface superchunk = 16x16 chunks
	# dungeon superchunk = 32x32 chunks
	# map: 8x8 superchunks + 5 dungeon superchunks == 32256 bytes
	# currently we only read in surface superchunks
	MAP: while (1) {
		my ($buf, @schk);
		SUPERCHUNK: while (1) {
			last MAP unless read MAP, $buf, 3;
			# Obtain 2 tiles from 3 bytes (6 nibbles)
			my @bytes = unpack("C3", $buf);
			push @schk, (($bytes[1] & 0xf) << 8) | $bytes[0];
			push @schk, ($bytes[2] << 4) | ($bytes[1] >> 4);
		#	printf "%03x " x @schk, @schk; print "\n";
			last SUPERCHUNK if @schk == 256;
		}
		push @map, \@schk;
		last if @map == 64;  # Read in 64 surface superchunks; the rest are dungeon superchunks
	}
	close(MAP);
}

# Convert (superchunk-x, schunk-y, chunk-x, chunk-y, tile-x, tile-y) coordinates to (world-x, world-y).
sub chunk_to_world {
	return ($_[0] * 128 + $_[2] * 8 + $_[4], 
	        $_[1] * 128 + $_[3] * 8 + $_[5]);
}

# Convert (world-x, world-y) to (superchunk-x, schunk-y, chunk-x, chunk-y, tile-x, tile-y).
# World coordinates are taken mod 1024, so you don't have to wrap, though
# it's probably required elsewhere.
sub world_to_chunk {
	my ($x, $y) = @_;
	return ($x >> 7 & 0x7, $y >> 7 & 0x7,  
	        $x >> 3 & 0xF, $y >> 3 & 0xF,
	        $x      & 0x7, $y      & 0x7);
}

# Return and/or set the map tile number at the specified world coordinates.
# This subroutine can be assigned to as an lvalue.
sub maptile_at : lvalue {
	my ($wx, $wy) = (shift, shift);
	my ($scx, $scy, $cx, $cy, $tx, $ty) = world_to_chunk($wx, $wy);
	my $schunk = $scx + $scy * 8;   # 8x8 superchunks / map
	my $chunk = $map[$schunk][$cx + $cy * 16];     # 16x16 chunks / schunk
	$chunks[$chunk][$tx + $ty * 8]                 # 8x8 tiles / chunk
}

sub write {
	local $/;
	rename $filename{chunks}, $filename{chunks} . ".old";
	open(CHUNKS, ">files/chunks") or die "$!";
	print CHUNKS pack("C*", @$_) foreach @chunks;
	close (CHUNKS);
}	

1;
