# $Id: Obj.pm,v 1.34 2003/02/12 23:58:53 jim Exp $

package U6::Obj;
BEGIN {
	use Exporter;
	our ($VERSION, @ISA, @EXPORT, @EXPORT_OK);
	$VERSION = do { my @r = (q$Revision: 1.34 $ =~ /\d+/g); sprintf "%d."."%02d" x $#r, @r };
	@ISA = qw(Exporter);
	@EXPORT    = qw();
	@EXPORT_OK = qw(@objblk print_obj_info map_type_to_tile is_unknown STATUS_UNKNOWN objects_at add_object_at remove_object_at remove_container default_object add_container);
}

use strict;
use Data::Dumper;
use U6::Look qw(get_obj_name);

our @objblk;
our %changed;

my $dir = "files/savegame/";
my $prefix = "objblk";
my $buf;
my @basetile;

my $Egg = 335;

my %filename = ( objblk => "$dir/$prefix" );

sub init {
	local $/;
	open BASETILE, "files/basetile" or die "$!";
	@basetile = unpack "S*", <BASETILE>;
	# die "incorrect number of basetiles: ", scalar @basetile, "\n" unless @basetile == 1024;
	close BASETILE;

	@objblk = ();
	undef %changed;
	for my $j ('a' .. 'h') {
		for my $i ('a' .. 'h') {
			my $file = $filename{objblk} . $i . $j;
			my $num_objs;
			my @objects = ();
			my @objidx  = ();  # Is it faster to use undef?

			open OBJBLK, $file or die "$!";
			read OBJBLK, $num_objs, 2;
			$num_objs = unpack "S", $num_objs;
			# print "$file: ", $num_objs, " objects\n";

			while (1) {
				last unless read OBJBLK, $buf, 8;
				my ($status, $h, $d1, $d2, $type, $quantity, $quality) = unpack "CCCCSCC", $buf;
				my ($x, $y) = ((($d1 & 0x03) << 8) | $h, (($d2 & 0xf) << 6) | (($d1 & 0xFC) >> 2));
				# This may not all be z coordinate, but we do need to save it.
				my $z = ($d2 & 0xF0) >> 4;
				my %object = ( status => $status, h => $h, d1 => $d1, d2 => $d2, x => $x, y => $y, 
							z => $z, type => $type, tile => map_type_to_tile($type), quantity => $quantity, quality => $quality );
				if(in_container(%object)) {
					# x-coordinate stores index of container.  This is a bit more
					# lenient than the game code.
					my $idx = $object{x};  
					my $obj = $objidx[$idx];
#					print "** ", get_obj_name($object{tile}), " contained by object index ", $idx, "\n";
#					print Dumper($obj);
#					print "   ** name ", get_obj_name($obj->{tile}), "\n";
					push @{ $obj->{contains} }, \%object;  # Store object in container (in file order)
				} else {
					push @objects, \%object;
				}
				push @objidx, \%object;  # Push all object (refs) onto the index array, so
				                         # we can refer back to containers correctly.
			}
			close OBJBLK;
			# Reverse @objects so that they appear in drawing order.
			push @objblk, [ reverse @objects ];
		}
	}
	return 1;
}

sub block_num_to_id {
	my $num = shift;
	my $x = $num % 8;
	my $y = $num / 8;
	my @ids = ('a' .. 'h');
	return $ids[$x] . $ids[$y];
}

sub block_id_to_num {
	my $id = shift;
	die "invalid block id $id" if $id !~ /^([a-h])([a-h])$/;
	my $x = ord($1) - ord('a');
	my $y = ord($2) - ord('a');
	my $num = $x + $y * 8;
	die if $num > 63;
	return $num;
}

sub world_to_block {
	my ($x, $y) = @_;
	return ($x >> 7) + ($y >> 7) * 8;
}

# Returns a list of objects at the specified coordinates, in drawing order
# (back to front--painter's algorithm).  Also return a reference to object
# block, and index into the block of the next object (i.e. the topmost object + 1).

sub objects_at {
	my ($wx, $wy) = @_;
	my @objs = ();
	my $block_num = world_to_block($wx, $wy);
	my ($index, $overshoot) = (0, 0);
	my $block = $objblk[$block_num];

	for (@$block) {
		# Objects are stored right-to-left, then bottom-to-top.

		$overshoot++, next if is_unknown(%$_);
		last if $_->{y} < $wy;
		last if $_->{y} == $wy and $_->{x} < $wx;
		# Whenever an object we can handle is found, zero out the overshoot.
		$overshoot = 0;
		next if $_->{y} > $wy;
		next if $_->{x} > $wx;
		# We've found a matching object.
		push @objs, $_;
	} continue {
		$index++;
	}
	# Subtract any overshoot (objects in party, not currently handled, for example).
	$index -= $overshoot;
	return ($block, $index, @objs);
}

# Find index into objblk of specified object.  Note: does -not- return
# object + 1, as objects_at does.
sub object_at {
	my ($wx, $wy, $obj) = @_;
	return undef unless $obj->{x} == $wx and $obj->{y} == $wy;

	my $block_num = world_to_block($wx, $wy);
	my $block = $objblk[$block_num];
	for (my $i = 0; $i < @$block; $i++) {
		return ($block, $i) if $block->[$i] == $obj;
	}
	return undef;
}

# Add specified object at world coordinates.
# If $match specified, will add as a left sibling to $match;
# fails if $match not found.  If $match not provided, will
# add as topmost (rightmost) object, which can be thought
# of as the left sibling to 'nil'.
sub add_object_at {
	my ($wx, $wy, $obj, $match) = @_;
	my ($objs, $index);
	if ($match) {
		($objs, $index) = object_at($wx, $wy, $match);
		return undef unless defined $index;
	} else {
		# When searching for the last object, we have
		# to use a special algorithm.
		($objs, $index) = objects_at($wx, $wy);
	}
	$obj->{x} = $wx;
	$obj->{y} = $wy;
	$obj->{status} &= 0x01;  # Not a container; only allow 0wned flag.
	splice(@$objs, $index, 0, $obj);	
	# Note that this block has changed.  
	my $block_num = world_to_block($wx, $wy);
	$changed{$block_num}++;  # could this overflow?
}

# Remove specified object from world coordinates.
# If no $match is provided, remove topmost object.
sub remove_object_at {
	my ($wx, $wy, $match) = @_;
	my ($obj, $objblk, $index, @objs);
	my $found;

	if ($match) {
		($objblk, $index) = object_at($wx, $wy, $match);
		$found = defined $index;
		$index++;   # Increase index here to compensate
		            # for the subtraction below.
	} else {
		($objblk, $index, @objs) = objects_at($wx, $wy);
		$found = @objs;
	}
	# Only remove object if there -is- an object to be removed.
	# $index points to the object -after- the topmost object.
	if ($found) {
		$obj = splice(@$objblk, $index - 1, 1);
		my $block_num = world_to_block($wx, $wy);
		$changed{$block_num}++;  # could this overflow?
	}
	return $obj;
}

# Add $object to container as left sibling of $match;
# consequently, to the end if $match is undef.
# The only reason we need the coordinates of the container
# is so that we can flag the block as changed.  Definitely
# kludgey.
sub add_container {
	my $wx = shift;
	my $wy = shift;
	my $container = shift;
	my $object = shift;
	my $match = shift;
	my $contains;
	my $i = 0;

	# Force an update 
	my $block_num = world_to_block($wx, $wy);
	$changed{$block_num}++;  # could this overflow?

	if (not $match) {
		# No match, add to end.  Create a contained-objects
		# array if not already present.
		$container->{contains} ||= [];
		$contains = $container->{contains};
 		$object->{status} &= 0x01;  # Container, and only allow 0wned flag.
		$object->{status} |= 0x08;
		$object->{y} = 0;           # Arbitrary y coordinate
		push(@$contains, $object);
		return 1;
	}

	$contains = $container->{contains};
	for ($i = 0; $i < @$contains; $i++) {
		if ($contains->[$i] == $match) {
			splice(@$contains, $i, 0, $object);
			$object->{status} &= 0x01;
			$object->{status} |= 0x08;
			$object->{y} = 0;
			return 1;
		}
	}
	return undef;
}

# This is getting out of hand
sub remove_container {
	my $wx = shift;
	my $wy = shift;
	my $container = shift;
	my $object = shift;
	# Force an update 
	my $block_num = world_to_block($wx, $wy);
	$changed{$block_num}++;  # could this overflow?
	$container->{contains} = [ grep { $_ != $object } @{ $container->{contains} } ];
}

# Write specified block number out (even if it hasn't been changed)
sub write {
	my $block_num = shift;
	my $block = $objblk[$block_num];
	my $output;
	my $index = 0;
	my $blockid = block_num_to_id($block_num);
	$index = write_objects($index, \$output, [ reverse @$block ]);
	# Prepend number of objects.
	substr($output, 0, 0) = pack "S", $index;
	my $file = $filename{objblk} . $blockid;
	rename $file, $file . ".old";
	open OBJBLK, ">", $file or die "$!";
	print OBJBLK $output;
	close OBJBLK;
	delete $changed{$block_num};
	return 1;
}

# Write all blocks that have been changed, out
sub write_changes {
	for (keys %changed) {
		print "- writing block ", block_num_to_id($_), " ($_)\n";
		&write($_);
	}
	return 1;
}

# args: $index:  next available object slot
#       $output: reference to output buffer
#       $objs:   reference to object array
# Returns next available object index into the objblk buffer (i.e., number of objects).
sub write_objects {
	my $index  = shift;
	my $output = shift;
	my $objs   = shift;
	my $index_container = $index - 1;
	for (@$objs) {
		my $status = $_->{status};
		my $type   = $_->{type};
		# Objects in containers have an x coordinate corresponding
		# to the container's index into the objblk file.
		my $x = in_container(%$_) ? $index_container : $_->{x};
		# The arbitrary (0) y-coordinate for containers seems to work,
		# but I'll leave it out for now.
#		my $y = in_container(%$_) ? 0 : $_->{y};
		my $y = $_->{y};
		my $z = $_->{z};
		my $h  = $x & 0xFF;
		my $d1 = ($x >> 8 & 0x03) | ($y << 2 & 0xfc);
		my $d2 = ($y >> 6 & 0x0F) | ($z << 4 & 0xF0);
		my $qty = $_->{quantity};
		my $obj = pack "CCCCSCC", $_->{status}, $h, $d1, $d2,
		                          $_->{type}, $_->{quantity}, $_->{quality};
		$$output .= $obj;
		$index++;
		my $contains = $_->{contains};
		if (defined $contains) {
			# The actual container index will be one less than the $index 
			# value we pass in, since $index refers to the next available slot.
			$index = write_objects($index, $output, $contains);
		}
	}
	return $index;
}

# Debugging: print object information
sub print_obj_info {
	my %obj = @_;
	printf "status: %02x x: %03x y: %03x h: %02x d1: %02x d2: %02x\n", $obj{status}, $obj{x}, $obj{y}, $obj{h}, $obj{d1}, $obj{d2};
	printf "type: %04x (%04x frm %d) quantity: %02x quality: %02x\n", $obj{type}, $obj{type} & 0x3ff, $obj{type} >> 10, $obj{quantity}, $obj{quality};
	printf "name: %s\n", get_obj_name(map_type_to_tile($obj{type}));
}

# Return desired frame of object.  Note that this will cause overhead when looking up the name
# of an object with several frames (since that requires the last frame).
sub map_type_to_tile {
	my $type = shift;
	return $basetile[($type & 0x3ff)] + ($type >> 10);
}

# Faster version of U6::Look::get_obj_name.  Relies on the fact that basetile
# always gives you the first frame of an object, and the look data is linked
# to the last frame.  So we get the basetile of the next object, and subtract 1.
# This only works for objects, not fixed maptiles.
# XXX This is invalid.  Do not use.  Look data can be linked to another object,
# not just next frame.
sub get_name {
	my $type = shift;
	my $objnum = $type & 0x3ff;  # Get object number (masking off frame)
	$objnum++;
	my $tile = $basetile[$objnum] - 1;
	return U6::Look::get_name($tile);
}

# Get a (long) list of object names.
sub get_all_names {
	map { get_obj_name($_) } @basetile;
}

sub num_frames {
	use U6::Tile qw(@maptiles);
	my $type = shift;
	$type &= 0x3ff;
	# Number of frames is difference between consecutive basetiles. 
	my $next = ($type + 1 == @basetile) ? @maptiles
	                                     : $basetile[$type + 1];
	return $next - $basetile[$type];
}

sub is_egg {
	my %obj = @_;
	return defined $obj{type} ? $obj{type} == $Egg : undef;
}

sub in_party {
	my %obj = @_;
	return defined $obj{status} ? $obj{status} & 0x10 : undef;
}

# Calling this in a tight loop is very expensive; we need to have
# it inlined, but to do that we have to make it a constant.
sub is_unknown {
	my %obj = @_;
	# Don't understand (or don't handle) the top four bits
	# or bits 2 and 1.
	return defined $obj{status} ? $obj{status} & 0xF6 : undef;
}

# See above.
sub STATUS_UNKNOWN() { 0xF6 }

sub in_container {
	my %obj = @_;
	# Assume objects are contained if their low nibble is 0x08 or 0x09, 
	# and they are not in the party (for now).
	return defined $obj{status} ? (($obj{status} & 0x1E) == 0x08) : undef;
}

sub we_0wn {
	my %obj = @_;
	return defined $obj{status} ? $obj{status} & 0x01 : undef;
}

sub clone {
	my $obj = shift;
	my %object = %$obj;
#	delete $object{contains};

	# Clone contained objects as well
	$object{contains} = [ map { clone($_) } @{ $object{contains} } ];
#	foreach my $cnt (@{ $object{contains} }) {
#		$cnt = clone($cnt);
#	}
		
	return \%object;
}

sub default_object {
	my ($x, $y, $z) = @_;
	my %new_object;
	$new_object{x} = $x;
	$new_object{y} = $y;
	$new_object{z} = $z;
	$new_object{type} = 1;
	$new_object{tile} = map_type_to_tile($new_object{type});
	$new_object{quantity} = 0;
	$new_object{quality}  = 0;
	$new_object{status} = 0x01;  # Make sure we own it.
	return \%new_object;
}

