package U6::ObjEdit;

use strict;
use Data::Dumper;
use U6::Obj qw(map_type_to_tile);
use Gtk::HexEntry;

sub create {
	my $obclass = shift;
	my $class = ref($obclass) || $obclass;
	my $self = {
		win   => new Gtk::Dialog(),
		stackedit => shift,
		adj   => undef,
		spin  => undef,
		object => undef,
	};
	bless $self, $class;

	my $win = $self->{win};
	$win->set_policy( 1, 1, 0);
	$win->set_title( "Object Editor" );
	$win->signal_connect( 'delete_event', sub { $_[0]->hide(); return 1; } );
	$win->set_usize( 300, 0 );
	
	my $vbox_coords = new Gtk::VBox( 0, 0 );
	my $vbox_qty    = new Gtk::VBox( 0, 0 );
	my (%spin, %adj, %hbox);
	$adj{x}  = new Gtk::Adjustment( 0, 0, 1023, 1, 64, 0 );
	$adj{y}  = new Gtk::Adjustment( 0, 0, 1023, 1, 64, 0 );
	$adj{z}  = new Gtk::Adjustment( 0, 0,    4, 1,  1, 0 );
	$adj{Status}  = new Gtk::Adjustment( 0, 0, 255, 1,  1, 0 );
	$adj{Status}->signal_connect('value_changed', sub { $self->sane_status() } );
			
	for ("x", "y", "z", "Status") {
		$spin{$_} = new Gtk::HexEntry( $adj{$_} );
		$spin{$_}->set_usize(30, 0);
		$hbox{$_} = new Gtk::HBox( 0, 0 );
		$hbox{$_}->pack_start(new Gtk::Label($_), 1, 1, 0);
		$hbox{$_}->pack_start($spin{$_}, 0, 0, 0);
		$vbox_coords->pack_start($hbox{$_}, 1, 1, 0);
		$adj{$_}->signal_connect('value_changed', sub { $self->update() } );
		$spin{$_}->signal_connect('activate', sub { $self->apply() } );
	}
	for ("Quality", "Quantity", "Frame") {
		$adj{$_}  = new Gtk::Adjustment( 0, 0, 255, 1, 16, 0 );
		$spin{$_} = new Gtk::SpinButton( $adj{$_}, 1.0, 0 );
		$hbox{$_} = new Gtk::HBox( 0, 0 );
		$hbox{$_}->pack_start(new Gtk::Label($_), 1, 1, 0);
		$hbox{$_}->pack_start($spin{$_}, 0, 0, 0);
		$vbox_qty->pack_start($hbox{$_}, 0, 0, 0);
		$adj{$_}->signal_connect('value_changed', sub { $self->update() } );
		$spin{$_}->signal_connect('activate', sub { $self->apply() } );
	}	
	$self->{adj}  = \%adj;
	$self->{spin} = \%spin;

	# Set up combo box containing object names.
	my $combo = new Gtk::Combo();
	$self->{combo} = $combo;
	my @items = U6::Obj::get_all_names();
	$combo->set_popdown_strings(map { sprintf "[%04x] %s", $_, $items[$_] } (0 .. $#items) );
	$combo->set_use_arrows_always( 1 );
	$combo->set_value_in_list( 1, 0 );
	# Inexcusable GTK crap.  You can't receive the list index clicked on, you have
	# to linearly search the list, and it matches on the string!  That does not
	# bode well for the wxWindows implementation either.
	# Actually, it appears that the list is rescanned and the first matching item
	# is selected, whenever you select.  So you can't select non-unique items.
	# Workaround: store index as user data in each listitem (for speed) and
	# prefix each item with the item number.
	$combo->entry->signal_connect('changed', sub { 
			my $objnum = $self->objnum(); 
			$self->{adj}->{Frame}->upper(U6::Obj::num_frames($objnum) - 1);
			$self->{adj}->{Frame}->set_value(0);
		}
	);

	$combo->entry->signal_connect('changed', sub { $self->update() } );

	my $index = 0;
	for ($combo->list->children) {
		my $copy = $index;
		$_->set_user_data(\$copy);
		$index++;
	}

	$win->vbox->pack_start( $combo, 0, 0, 0 );
	my $hbox_hold = new Gtk::HBox( 0, 0 );
	$hbox_hold->pack_start($vbox_coords, 1, 1, 0 );
	$hbox_hold->pack_start($vbox_qty, 1, 1, 0 );
	$win->vbox->pack_start( $hbox_hold, 0, 0, 5 );

	my $apply = new Gtk::Button("Apply");
	$apply->signal_connect('clicked', sub { $self->apply() } );
	my $auto = new Gtk::CheckButton("Auto");
	$auto->set_active( 1 );
	$self->{auto} = $auto;
	$win->action_area->pack_start( $auto, 1, 1, 10 );
	$win->action_area->pack_start( $apply, 1, 1, 10 );

#	$win->vbox->pack_start( $vbox_coords, 0, 0, 0 );
#	$win->vbox->pack_start( $vbox_qty, 0, 0, 0 );

	return $self;
}

sub show {
	$_[0]->{win}->show_all();
}

# Yuck, the type/frame conversion is hardcoded here.
sub object {
	my $self = shift;
	my $obj = shift;
	# Don't bother updating if this object is already on the screen.
#	return if $self->{object} and $self->{object} == $obj;

	$self->{object} = $obj;
	$self->{frozen} = 1;
	# Set coordinates
	for ("x", "y", "z", "Status") {
		$self->{spin}->{$_}->set_value($obj->{lc($_)});
	}
	$self->{combo}->list->select_item($obj->{type} & 0x3ff);
	$self->{spin}->{Quality}->set_value($obj->{quality});
	$self->{spin}->{Quantity}->set_value($obj->{quantity});
	$self->{adj}->{Frame}->upper(U6::Obj::num_frames($obj->{type}) - 1);
	$self->{spin}->{Frame}->set_value($obj->{type} >> 10);
	$self->{frozen} = 0;
	return 1;
}

sub update {
	my $self = shift;
	my $w = shift;

	my $obj = {};
	if ($self->{auto}->get_active() and not $self->{frozen}) { 
		$self->apply();
		$obj = $self->{object};
	} else {
		# The point of this "update" is simply to generate a valid
		# temporary object based on the current parameters, so 
		# we can set the window title / change whatever else.
		$self->update_object($obj);
	}
	$self->{win}->set_title( sprintf("Object (%03x, %03x) %s", 
	      $self->{object}->{x}, $self->{object}->{y}, ::get_parsed_name($obj)));
}

sub update_object {
	my $self = shift;
	my $obj  = shift;
	my $s = $self->{spin};
	$obj->{quality}  = $s->{Quality}->get_value_as_int;
	$obj->{quantity} = $s->{Quantity}->get_value_as_int;
	$obj->{status}   = $s->{Status}->value;
	$obj->{type}     = $self->objnum() | ($s->{Frame}->get_value_as_int << 10);
	$obj->{tile}     = map_type_to_tile($obj->{type});
}

sub apply {
	my $self = shift;
	my $obj  = $self->{object};
	$self->update_object($obj);
	# A signal/slot strategy might serve well here
	$self->{frozen} = 1;
	$self->{stackedit}->rebuild();
	$self->{frozen} = 0;
}

sub objnum {
	return ${ $_[0]->{combo}->list->selection->get_user_data() };
}

sub sane_status {
	my $self = shift;
	my $val = $self->{spin}->{Status}->value;
	my $old_val = $self->{object}->{status};
	# Only allow changes to bit 0 (0wned).
	$val &= 0x01;
	$val |= $old_val & ~0x01;
	$self->{spin}->{Status}->set_value($val);
}

1;
