# $Id: Pal.pm,v 1.1 2003/02/08 05:29:57 jim Exp $
package U6::Pal;

BEGIN {
	use Exporter;
	our ($VERSION, @ISA, @EXPORT, @EXPORT_OK);
	$VERSION = do { my @r = (q$Revision: 1.1 $ =~ /\d+/g); sprintf "%d."."%02d" x $#r, @r };
	@ISA = qw(Exporter);
	@EXPORT    = qw();
	@EXPORT_OK = qw(&pal &rotate_palette);
}

use strict;

sub new {
	my $type = shift;
	my $class = ref($type) || $type;
	my $self = {
		pal => undef,
	};
	bless $self, $class;

	open(PAL, "files/u6pal") or die "$!";
	my @pal;
	# File contains an extra 256 bytes (with unknown function)
	for (0 .. 255) {
		my $palentry;
		read PAL, $palentry, 3;
		my ($r, $g, $b) = unpack "C3", $palentry;
		# VGA palette is only 6-bits per color
		push @pal, ($r << 18) | ($g << 10) | ($b << 2);
	}
	$pal[255] = 0;  # faked transparent, for drawing transparent tiles with opaque functions
	close PAL;

	$self->{pal} = \@pal;
	return $self;
}

sub pal {
	return @{ $_[0]->{pal} };
}

# Arguments: cycle-length, first interval start, second interval start, ...
# This is actually a generic partial array rotation function, with array hardcoded.
sub rotate_palette {
	my $pal = shift->{pal};
	# Palettes rotate to the left (according to u6tech.txt)
	my $len = (shift) - 1;
	splice(@$pal, $_ + $len, 0, splice(@$pal, $_, 1)) foreach @_;
}

1;
