package U6::StackEdit;

use strict;
use Data::Dumper;
use U6::ObjEdit;
use U6::Obj qw(remove_object_at objects_at remove_container add_object_at default_object add_container);

sub create {
	my $obclass = shift;
	my $class = ref($obclass) || $obclass;
	my $self = {
		win   => new Gtk::Dialog(),
		swin  => new Gtk::ScrolledWindow(undef, undef),
		ctree => new Gtk::CTree(1, 0),
		wx    => undef,
		wy    => undef,
		wz    => 0
	};
	$self->{objedit} = create U6::ObjEdit($self);
	bless $self, $class;

	my $win = $self->{win};
	$win->set_policy( 1, 1, 0);
	$win->set_title( "Stack Editor" );
	$win->signal_connect( 'delete_event', sub { $_[0]->hide(); return 1; } );
	
	my $swin = $self->{swin};
	$swin->set_policy( 'always', 'automatic' );  # scrollbar policy

	my $ctree = $self->{ctree};
	# FIXME This is a stupid call; column width should be automatically determined.
	$ctree->set_column_width(0, 200);
	# FIXME Also a stupid call; but currently necessary so grandma make window bigger than postage stamp.
	$ctree->set_usize( 300, 250 );
	$ctree->set_reorderable( 1 );
	$ctree->signal_connect('tree_select_row', sub { $self->select_row(@_) } );
	$ctree->signal_connect('tree_move', sub { $self->apres_move(@_) } );

	$swin->add($ctree);

	my $add = new Gtk::Button("Add");
	$add->signal_connect('clicked', sub { $self->add() } );
	my $dup = new Gtk::Button("Dup");
	$dup->signal_connect('clicked', sub { $self->dup() } );

	my $remove = new Gtk::Button("Remove");
	$remove->signal_connect('clicked', sub { $self->remove() } );
	
	$win->vbox->pack_start( $swin, 1, 1, 0 );
	$win->action_area->pack_start( $add, 1, 1, 5 );
	$win->action_area->pack_start( $dup, 1, 1, 5 );
	$win->action_area->pack_start( $remove, 1, 1, 5 );

	return $self;
}

sub select_row {
	my $self = shift;
	my ($w, $node, $col) = @_;
#	my ($text) = $w->get_node_info($node);
#	print "stack: selected $text\n";
	my $obj = $w->node_get_row_data($node);
	$self->{objedit}->object($obj);
	$self->{objedit}->show();
}

sub show {
	$_[0]->{win}->show_all();
}

sub build {
	my $self = shift;
	my ($wx, $wy, $objs, $select) = @_;
	my $ctree = $self->{ctree};
	$select = 1 unless defined $select;  # Select last item
	
	$ctree->freeze;
	$ctree->clear;

	$self->recursive_add(undef, $objs, $select);	
		
	$ctree->thaw;
	$self->{win}->set_title( sprintf("Stack (%03x, %03x)", $wx, $wy) );
	$self->{objs} = $objs;
	$self->{wx} = $wx;
	$self->{wy} = $wy;
	return 1;
}


sub add {
	my $self = shift;
	my $new_object = shift;
	my $node = $self->{ctree}->selection;
	my ($parent, $sibling, $sibling_object);
	$new_object ||= default_object($self->{wx}, $self->{wy}, $self->{wz});

	$parent = $node->row->parent if $node;
	$sibling = $node->row->sibling if $node;
	# We need to add to the right sibling object of the selection, 
	# because the add functions will add us as a left sibling.
	# This will splice us in between them.
	$sibling_object = $self->{ctree}->node_get_row_data( $sibling ) if $sibling;
	if ($parent) {
		my $container = $self->{ctree}->node_get_row_data( $node->row->parent );
		add_container($self->{wx}, $self->{wy}, $container, $new_object, $sibling_object);
	} else {
		# Will be added on top, if no object selected.
		add_object_at($self->{wx}, $self->{wy}, $new_object, $sibling_object);
	}
	my (undef, undef, @objs) = objects_at($self->{wx}, $self->{wy});
	$self->build($self->{wx}, $self->{wy}, \@objs, 0);
	$self->select_node_by_data($new_object);
	::draw_map_viewport();
}

sub remove {
	my $self = shift;
	my $node = $self->{ctree}->selection;
	my $object = $self->selected_object();
	return unless $object;

	my ($parent, $sibling, $parent_obj, $sibling_obj);
	$parent = $node->row->parent;
	$sibling = $node->row->sibling;
	$parent_obj = $self->{ctree}->node_get_row_data( $parent ) if $parent;
	$sibling_obj = $self->{ctree}->node_get_row_data( $sibling ) if $sibling;
	if ($parent) {
		my $container = $parent_obj;
		remove_container($self->{wx}, $self->{wy}, $container, $object);
	} else {
		remove_object_at($self->{wx}, $self->{wy}, $object);
	}
	my (undef, undef, @objs) = objects_at($self->{wx}, $self->{wy});
	$self->build($self->{wx}, $self->{wy}, \@objs, 1);

	# Try to select a reasonable item.  This isn't ideal, because $sibling_obj
	# will only give us the right child, which doesn't exist for the last item.
	if ($sibling_obj) {
		$self->select_node_by_data($sibling_obj);
	} elsif ($parent_obj) {
		$self->select_node_by_data($parent_obj);
	}
	::draw_map_viewport();
}

sub dup {
	my $self = shift;
	my $obj = $self->selected_object();
	return undef unless $obj;

	$self->add(U6::Obj::clone($obj));
}

sub selected_object {
	my $self = shift;
	my $object;
	my $node = $self->{ctree}->selection;
	$object = $self->{ctree}->node_get_row_data( $node ) if $node;
	return $object;
}

sub rebuild {
	my $self = shift;
	# Don't select last item on rebuild, because it could trigger
	# recursion with ObjEdit. (Actually, this doesn't appear to
	# be the case, but there's no reason to select it.)
	my $last_obj;
	$last_obj = $self->selected_object();

	my $objs = $self->{objs};
	if (not defined $objs) {
		# Rebuild our object database from scratch if undefined.
		my (undef, undef, @objs) = objects_at($self->{wx}, $self->{wy});
		$objs = \@objs;
	}
	$self->build($self->{wx}, $self->{wy}, $objs, 0);

	# The tree does not keep node values constant (try selecting the
	# same node multiple times, the node ref will be different).
	# Instead, we search the tree for our associated object reference,
	# and select it if found.
 	$self->select_node_by_data($last_obj);

	# Is this necessary?
	::draw_map_viewport();
}

sub recursive_add {
	my $self   = shift;
	my $parent = shift;
	my $objs   = shift;
	my $select = shift;
	
	my $this;
	for my $obj (@$objs) {
		$this = $self->{ctree}->insert_node($parent, undef, 
		                  [ ::get_parsed_name($obj) ], 5, 
		                    undef, undef, undef, undef, 0, 1);
		$self->{ctree}->node_set_row_data($this, $obj);
		$self->recursive_add($this, $obj->{contains}, 0) if defined $obj->{contains};
	}
	# Select object on top
	$self->{ctree}->select($this) if $this and $select;
	return 1;
}


sub select_node_by_data {
	my $self = shift;
	my $match = shift;
	return undef unless $match;
	# I can't get the closure to read its arguments, so
	# I have it read my lexical.
	$self->{ctree}->post_recursive( undef, sub {
		my ($c, $n) = @_;
		my $obj  = $c->node_get_row_data( $n );
		if ($obj == $match) {
			$c->select($n);   # Not sure how to terminate early.
		}
	} );
}

# Cribbed from Gtk-Perl's test.pl.
sub apres_move {
	my $self = shift;
	my ($ctree, $child, $new_parent, $sibling) = @_;
	my ($source) = $ctree->get_node_info($child);
	my $old_parent = $child->row->parent;
	my ($source_p) = $ctree->get_node_info($old_parent) if $old_parent;
	my ($dest_p) = $ctree->get_node_info($new_parent) if $new_parent;
	my ($dest_sib) = $ctree->get_node_info($sibling) if $sibling;

	$source_p ||= 'surface';
	$dest_p   ||= 'surface';
	$dest_sib ||= 'nil';
	
	print "- moving $source from $source_p to $dest_p, before $dest_sib\n";

	my $object = $ctree->node_get_row_data( $child );

	if ($old_parent) {
		my $container = $ctree->node_get_row_data( $old_parent );
		remove_container($self->{wx}, $self->{wy}, $container, $object);
	} else {
		remove_object_at($self->{wx}, $self->{wy}, $object);
	}

	my $sibling_object;
	$sibling_object = $ctree->node_get_row_data( $sibling ) if $sibling;

	if ($new_parent) {
		my $container = $ctree->node_get_row_data( $new_parent );
		add_container($self->{wx}, $self->{wy}, $container, $object, $sibling_object);
	} else {
		add_object_at($self->{wx}, $self->{wy}, $object, $sibling_object);
	}	

	# Don't bother rebuilding; we should be accurate enough, and
	# it looks nicer.  We do update objedit manually since the 
	# move may have changed data and the selection hasn't changed.
	$self->{objedit}->object($object);
	# Also, undef the record of our objects so that the next call
	# to rebuilt() will rebuild from scratch.
	undef $self->{objs};
	# Also, update the viewport ;)
	::draw_map_viewport();

	return 1;
}
	

1;
