#!/usr/bin/perl -w
# $Id: Tile.pm,v 1.4 2003/02/09 06:24:24 jim Exp $

package U6::Tile;
BEGIN {
	use Exporter;
	our ($VERSION, @ISA, @EXPORT, @EXPORT_OK);
	$VERSION = do { my @r = (q$Revision: 1.4 $ =~ /\d+/g); sprintf "%d."."%02d" x $#r, @r };
	@ISA = qw(Exporter);
	@EXPORT    = qw();
	@EXPORT_OK = qw(@masktypes @tileindx @maptiles %anim @tileflag update_animated_tiles tile_size);
}

use strict;

our (@masktypes, @tileindx, @maptiles, %anim, @tileflag);

sub init {
	local $/;
	# Load masktypes
	my $masktypes;
	open MASKTYPES, "uncompressed/masktype.vga" or die "$!";
	read MASKTYPES, $masktypes, 0x800;
	@masktypes = unpack("C*", $masktypes);
	close MASKTYPES;

	# Tile indices are not necessary unless we want to crosscheck maptiles validity
	open TILEINDX, "files/tileindx.vga" or die "$!";
	@tileindx = unpack("S*", <TILEINDX>);
	close TILEINDX;

	# Load maptiles
	open(MAPTILES, "uncompressed/alltiles.vga") or die "$!";
	for (my $i = 0;;$i++) {
		my $tile;
		last if $i >= @masktypes;  # shouldn't happen
		if ($masktypes[$i] == 0x0A) {
			my $buf;
			last unless read MAPTILES, $buf, 1;
			my $len = unpack("C", $buf);
			last unless read MAPTILES, $buf, $len * 16 - 1;
			$tile = "\377" x 256;
			my $tileidx = 0;
			while (length $buf) {
				my ($add, $pixels) = unpack "S C/a", $buf;
				last unless length $pixels;  # End of tile if block length is 0.
				my $actual_add = ($add % 160) 
							+ ($add >= 1760 ? 160 : 0);  # modulus fails for $add >= 1760 (kludge!!)
				$tileidx += $actual_add;
				substr($tile, $tileidx, length $pixels) = $pixels;
				$tileidx += length $pixels;
				substr($buf,0,(length $pixels) + 3,"");
			}
		} else {
			last unless read MAPTILES, $tile, 256; # Read one tile
		}
		push @maptiles, $tile;
	}
	close MAPTILES;

	# Load tile animation data
	open(ANIMDATA, "files/animdata") or die "$!";
	my $buf;
	# See u6tech.txt for structure.
	read ANIMDATA, $buf, 0x02;
	$anim{numtiles} = unpack "S", $buf;
	read ANIMDATA, $buf, 0x40;
	# tiles: tile numbers to animate
	$anim{tiles} = [ unpack "S32", $buf ];
	read ANIMDATA, $buf, 0x40;
	# first_frame: list of first-frame tile numbers, corresponding to 'tiles'
	$anim{first_frame} = [ unpack "S32", $buf ];
	read ANIMDATA, $buf, 0x20;
	$anim{and_masks} = [ unpack "C32", $buf ];
	read ANIMDATA, $buf, 0x20;
	$anim{shift_values} = [ unpack "C32", $buf ];
	close(ANIMDATA);

	# Load tileflags
	open(TILEFLAG, "files/tileflag") or die "$!";
	@tileflag = unpack "C*", <TILEFLAG>;
	close TILEFLAG;

}

sub update_animated_tiles {
	my $game_timer = shift;
	for my $i (0 .. $anim{numtiles} - 1) {  # I believe the -1 is correct, but am not positive.
		# Copy the appropriate animation frame (from the maptiles) into each animated tile.
		# Note: this does an actual copy [not a pointer swap] and is therefore slow.
		my $frame_num = ($game_timer & $anim{and_masks}[$i]) >> $anim{shift_values}[$i];
		my $tile  = $anim{tiles}[$i];
		my $frame = $anim{first_frame}[$i] + $frame_num;
		$maptiles[$tile] = $maptiles[$frame];
	}
}

# Return a 2-bit value representing tile size.
# If bit 1 set, object width  is 2.
# If bit 0 set, object height is 2.
sub tile_size {
	my $tile = shift;
	return ($tileflag[0x800 + $tile] >> 6) & 0x3;
}

sub article {
	my $tile = shift;
	my $article = ($tileflag[0x1400 + $tile] >> 6) & 0x3;
	return ("", "a", "an", "the")[$article];
}

1;
