package U6::TileEdit;
# $Id: TileEdit.pm,v 1.7 2003/02/13 06:30:14 jim Exp $

# FIXME The height of the window is not guaranteed to be a multiple of tile
# height, so we may cut the bottom tile off.
# FIXME uses global colormap $::cmap

use strict;
use U6::Tile qw(@maptiles);
use U6::Util;

sub TILES_X() { 16 }       # Number of tiles to show in the drawing area.
sub TILES_Y() { 20 }       # (Beginning) number of tiles

sub create {
	my $obclass = shift;
	my $class = ref($obclass) || $obclass;
	my $self = {
		win          => new Gtk::Window('toplevel'),
		drawing_area => new Gtk::DrawingArea(),
		tile_y       => 0,
		selected	 => 0,
	};
	bless $self, $class;

	my $win = $self->{win};
	$win->set_policy( 1, 1, 0 );
	$win->set_title( "Tile Viewer" );
	# Unlike the book editor, I'm hiding the window rather than destroying
	# it on close.  We'll see which turns out better.
	$win->signal_connect( 'delete_event', sub { $_[0]->hide(); return 1; } );

	# Create the tile viewer drawing area
	my $drawing_area  = $self->{drawing_area};
	my $height = TILES_Y * TILE_HEIGHT;
	my $width  = TILES_X * TILE_WIDTH;
	$drawing_area->size( $width, $height );

	my $vbox = new Gtk::VBox(0, 0);
	my $hbox = new Gtk::HBox(0, 0);
	# Gtk::Adjustment( $value, $lower, $upper, $step_increment, $page_increment, $page_size )
	$self->{adj}  = new Gtk::Adjustment( 0, 0, @maptiles / TILES_X, 1, 8, $height / TILE_HEIGHT );
	my $adj = $self->{adj};

	$adj->signal_connect('value_changed', sub {
			my $adj = shift;
			my $val = int $adj->value;
			# Redraw the viewport if we've moved a full tile.
			if ($self->{tile_y} != $val) {
				$self->{tile_y}  = $val;
				$self->{drawing_area}->queue_draw();
			}
		}
	);
	my $vscroll = new Gtk::VScrollbar( $adj );
	$win->add( $vbox );
	$vbox->pack_start($hbox, 1, 1, 0);
	$hbox->pack_start($drawing_area, 1, 1, 0);
	$hbox->pack_start($vscroll, 0, 0, 0);

	$drawing_area->signal_connect( 'configure_event',      sub { $self->tile_configure(@_) } );
	$drawing_area->signal_connect( 'expose_event',         sub { $self->draw_shapes(@_)    } );
	$drawing_area->signal_connect( 'button_press_event',   sub { $self->button_press(@_)   } );
	$drawing_area->signal_connect( 'button_release_event', sub { $self->button_release(@_) } );
	$drawing_area->set_events(['button_press_mask', 'button_release_mask', 'key_press_mask']);

	$self->select_tile(0);

	return $self;
}

# Tile area has been reconfigured (resized, typically).
# Update the page_size of the adjustment, and redraw.
sub tile_configure {
	my $self = shift;
	# Note: we ignore the widget passed by GTK. (Should be us, anyway!)
	my $height = $self->{drawing_area}->allocation->[3];
	$self->{adj}->page_size($height / TILE_HEIGHT);
	$self->draw_shapes();
}

sub draw_shapes
{
  	my $self = shift;
	my $draw = $self->{drawing_area};
	# Much of this bgcolor/white_gc stuff may be unnecessary.
    my $bgcolor = Gtk::Gdk::Color->parse_color( 'black' );
    $bgcolor = $draw->window->get_colormap()->color_alloc( $bgcolor );
    $draw->window->set_background( $bgcolor );

    my $drawable = $draw->window;
    my $white_gc = $draw->style->white_gc;

	my $x = 0; my $y = 0;
	for(my $i = $self->{tile_y} * TILES_X; $i < @maptiles; $i++) {
		my $tile = $maptiles[$i];
		$drawable->draw_indexed_image($white_gc, $x, $y,
		         16, 16, 'none', $tile, 16, $::cmap);
		$x += 16;
		$y += 16, $x = 0 if $x >= TILES_X * 16;
	}

	# Draw an outline around the selected tile; the math is
	# the reverse of that in button_press().
	my $tx =     $self->{selected} % TILES_X;
	my $ty = int $self->{selected} / TILES_X;
	$ty -= $self->{tile_y};
	($x, $y) = ($tx * TILE_WIDTH, $ty * TILE_HEIGHT);
	$drawable->draw_rectangle($white_gc, 0, $x, $y, 15, 15);
    return 1;
}

sub show {
	my $self = shift;
	$self->{win}->show_all();
}

# Selected tile number
sub selected {
	return shift->{selected};
}

sub select_tile {
	my $self = shift;
	my $tile = shift;
	$self->{selected} = $tile;
#	printf "- tile %d (0x%04x) selected\n", $tile, $tile;
	$self->{win}->set_title( sprintf("Tile Viewer (0x%03x %s)", $tile, U6::Obj::get_obj_name($tile)) );
}


# Button pressed inside the drawing area.
sub button_press {
	my $self = shift;
	my ($w, $e) = @_;
	my ($x, $y) = ($e->{x}, $e->{y});
	my $tx = int $x / TILE_WIDTH;
	my $ty = int $y / TILE_HEIGHT;
	$ty += $self->{tile_y};
	my $tile = $ty * TILES_X + $tx;
	$self->select_tile($tile);
	$self->draw_shapes();   # To update the selection
	return 1;
}

sub button_release {
	return 0;
}
	

=pod
=head1 Old palette drawing code
# Draw palette

#	my $width = 25; my $height = 512;
#	my @data;
#	for (my $y = 0; $y < $height / 2; $y++) {
#		push @data, ($y) x $width;
#		push @data, ($y) x $width;
#	}
#		
#	my $pixmap_data = pack "C*", @data;

#	$drawable->draw_indexed_image($white_gc, 20, 5, $width, $height, 'none', $pixmap_data, $width, $cmap); 
=cut

1;
